#!/usr/bin/env python

import sys, os, tempfile, subprocess, gc, cProfile
from time import time, sleep

import gtk, glib
from pylab import math, linspace, loadtxt

os.chdir(os.path.dirname(sys.argv[0]))
sys.path.insert(0, '..')

# A test function is a python generator function that may yield the
# values below to control the GUI and the testing:
#
wait_for_idle = -1 # wait until the last mypaint idle handler has finished
wait_for_gui = -2  # wait until all GUI updates are done, but don't wait for background tasks
start_measurement = -3
stop_measurement = -4

all_tests = {}

def run_test(testfunction, profile=None):
    """Run a single test
    testfunction must be a generator (using yield)
    """
    tst = testfunction()

    time_total = 0.0
    for res in tst:
        assert res == start_measurement
        def run_function_under_test():
            res = tst.next()
            assert res == stop_measurement
        t0 = time()
        if profile:
            profile.runcall(run_function_under_test)
        else:
            run_function_under_test()
        time_total += time() - t0

    if time_total:
        print 'result =', time_total
    else:
        pass # test did not make time measurements, it will print its own result (eg. memory)

def with_gui_setup(testfunction):
    """Add GUI setup code around a testfunction
    testfunction must be a generator (using yield)
    returns a new generator function
    """
    def testfunction_with_gui_setup():
        global app
        app = None
        tst = testfunction()
        # wait for the first command that requires the GUI
        while True:
            res = tst.next()
            if res in [wait_for_idle, wait_for_gui]:
                break
            else:
                yield res

        tempdir = tempfile.mkdtemp()
        try:
            from gui import application
            app = application.Application(datapath='..', confpath=tempdir, filenames=[])

            import gobject
            def signal_handler():
                global wait_for_signal_handler
                wait_for_signal_handler = False

            # fatal exceptions, please
            def excepthook(exctyp, value, tb):
                import traceback
                traceback.print_exception (exctyp, value, tb, None, sys.stderr)
                sys.exit(1)
            sys.excepthook = excepthook

            global wait_for_signal_handler
            wait_for_signal_handler = False
            while True:
                while not wait_for_signal_handler:
                    res = tst.next()
                    if res == wait_for_idle:
                        gobject.idle_add(signal_handler, priority=gobject.PRIORITY_LOW + 50)
                        wait_for_signal_handler = True
                    elif res == wait_for_gui:
                        gobject.idle_add(signal_handler, priority=gobject.PRIORITY_DEFAULT_IDLE - 1)
                        wait_for_signal_handler = True
                    elif res >= 0:
                        gobject.timeout_add(int(res*1000.0), signal_handler)
                        wait_for_signal_handler = True
                    else:
                        yield res
                gtk.main_iteration()
        finally:
            os.system('rm -rf ' + tempdir)

    return testfunction_with_gui_setup

def gui_test(f):
    "decorator to declare GUI test functions"
    all_tests[f.__name__] = with_gui_setup(f)
    return f

def nogui_test(f):
    "decorator for test functions that require no gui"
    all_tests[f.__name__] = f
    return f


@gui_test
def startup():
    yield start_measurement
    yield wait_for_idle
    yield stop_measurement

@gui_test
def paint():
    """
    Paint with a constant number of frames per recorded second.
    Not entirely realistic, but gives good and stable measurements.
    """
    yield wait_for_gui
    FPS = 30
    dw = app.drawWindow
    tdw = dw.tdw

    b = app.brushmanager.get_brush_by_name('redbrush')
    app.brushmanager.select_brush(b)

    dw.fullscreen_cb()
    yield wait_for_idle

    events = loadtxt('painting30sec.dat.gz')
    events = list(events)
    yield start_measurement
    t0 = time()
    t_old = 0.0
    t_last_redraw = 0.0
    for t, x, y, pressure in events:
        if t > t_last_redraw + 1.0/FPS:
            yield wait_for_gui
            t_last_redraw = t
        dtime = t - t_old
        t_old = t
        cr = tdw.get_model_coordinates_cairo_context()
        x, y = cr.device_to_user(x, y)
        dw.doc.stroke_to(dtime, x, y, pressure)
    yield stop_measurement

@gui_test
def paint_zoomed_out_5x():
    yield wait_for_idle
    dw = app.drawWindow
    for i in range(5):
        dw.zoom('ZoomOut')
    for res in paint():
        yield res

@gui_test
def layerpaint_nozoom():
    yield wait_for_idle
    app.filehandler.open_file('bigimage.ora')
    dw = app.drawWindow
    dw.doc.select_layer(len(dw.doc.layers)/2)
    for res in paint():
        yield res

@gui_test
def layerpaint_zoomed_out_5x():
    yield wait_for_idle
    dw = app.drawWindow
    app.filehandler.open_file('bigimage.ora')
    dw.tdw.scroll(800, 1000)
    dw.doc.select_layer(len(dw.doc.layers)/3)
    for i in range(5):
        dw.zoom('ZoomOut')
    for res in paint():
        yield res

@gui_test
def paint_rotated():
    yield wait_for_idle
    app.drawWindow.tdw.rotate(46.0/360*2*math.pi)
    for res in paint():
        yield res

@nogui_test
def load_ora():
    from lib import document
    d = document.Document()
    yield start_measurement
    d.load('bigimage.ora')
    yield stop_measurement

@nogui_test
def save_ora():
    from lib import document
    d = document.Document()
    d.load('bigimage.ora')
    yield start_measurement
    d.save('test_save.ora')
    yield stop_measurement

@nogui_test
def save_png():
    from lib import document
    d = document.Document()
    d.load('bigimage.ora')
    yield start_measurement
    d.save('test_save.png')
    yield stop_measurement

def scroll(zoom_func):
    dw = app.drawWindow
    dw.fullscreen_cb()
    app.filehandler.open_file('bigimage.ora')
    zoom_func()
    yield wait_for_idle

    yield start_measurement
    N = 20
    dx = linspace(-30, 30, N)
    dy = linspace(-10, 60, N)
    for i in xrange(N):
        dw.tdw.scroll(int(dx[i]), int(dy[i]))
        yield wait_for_idle
    yield stop_measurement

@gui_test
def scroll_nozoom():
    yield wait_for_idle
    def f(): pass
    for res in scroll(f):
        yield res

@gui_test
def scroll_zoomed_out_5x():
    yield wait_for_idle
    dw = app.drawWindow
    def f():
        for i in range(5):
            dw.zoom('ZoomOut')
    for res in scroll(f):
        yield res

@gui_test
def memory_zoomed_out_5x():
    yield wait_for_idle
    dw = app.drawWindow
    dw.fullscreen_cb()
    app.filehandler.open_file('bigimage.ora')
    for i in range(5):
        dw.zoom('ZoomOut')
    yield wait_for_idle
    dw.tdw.scroll(100, 120)
    yield wait_for_idle
    dw.tdw.scroll(-80, -500)
    yield wait_for_idle
    print 'result =', open('/proc/self/statm').read().split()[0]

if __name__ == '__main__':
    if len(sys.argv) == 4 and sys.argv[1] == 'SINGLE_TEST_RUN':
        func = all_tests[sys.argv[2]]
        if sys.argv[3] == 'NONE':
            run_test(func)
        else:
            # Not as useful as it could be, because most tests start the UI
            # The UI startup time dominates over the actual test execution time
            # Perhaps the tests should return a function that we can execute 
            # after setting everything up?
            profile = cProfile.Profile()
            run_test(func, profile)
            profile.dump_stats(sys.argv[3])
        sys.exit(0)

    from optparse import OptionParser
    parser = OptionParser('usage: %prog [options] [test1 test2 test3 ...]')
    parser.add_option('-a', '--all', action='store_true', default=False, 
                      help='run all tests')
    parser.add_option('-l', '--list', action='store_true', default=False,
                    help='list all available tests')
    parser.add_option('-c', '--count', metavar='N', type='int', default=3, 
                      help='number of repetitions (default: 3)')
    parser.add_option('-p', '--profile', metavar='PREFIX',
                    help='dump cProfile info to PREFIX_TESTNAME_N.pstats')
    options, tests = parser.parse_args()

    if options.list:
        for name in sorted(all_tests.keys()):
            print name
        sys.exit(0)

    if not tests:
        if options.all:
            tests = list(all_tests)
        else:
            parser.print_help()
            sys.exit(1)

    for t in tests:
        if t not in all_tests:
            print 'Unknown test:', t
            sys.exit(1)

    results = []
    for t in tests:
        result = []
        for i in range(options.count):
            print '---'
            print 'running test "%s" (run %d of %d)' % (t, i+1, options.count)
            print '---'
            # spawn a new process for each test, to ensure proper cleanup
            args = ['./test_performance.py', 'SINGLE_TEST_RUN', t, 'NONE']
            if options.profile:
                fname = '%s_%s_%d.pstats' % (options.profile, t, i)
                args[3] = fname
            child = subprocess.Popen(args, stdout=subprocess.PIPE)
            output, trash = child.communicate()
            if child.returncode != 0:
                print 'FAILED'
                break
            else:
                print output,
                try:
                    value = float(output.split('result = ')[-1].strip())
                except:
                    print 'FAILED to find result in test output.'
                    result = None
                    break
                else:
                    result.append(value)
        # some time to press ctrl-c
        sleep(1.0)
        if result is None:
            sleep(3.0)
        results.append(result)
    print
    print '=== DETAILS ==='
    print 'tests =', repr(tests)
    print 'results =', repr(results)
    print
    print '=== SUMMARY ==='
    fail=False
    for t, result in zip(tests, results):
        if not result:
            print t, 'FAILED'
            fail=True
        else:
            print '%s %.3f' % (t, min(result))
    if fail:
        sys.exit(1)
